// Jason Rohrer
// Color.h

/**
*
*	Color object for RadiosGL
*
*	Float components in RGB (0..1)
*	Color of emitted energy can be greater than 1
*
*
*	Created 9-5-99
*	Mods:
*		Jason Rohrer	11-6-99			Added 32-bit composite member
*		Jason Rohrer	11-13-99		Added printing functionality
*		Jason Rohrer	11-15-99		Added rebuildComposite function
*											Added weightColor function
*											Added getMax function
*
*/

#ifndef COLOR_INCLUDED
#define COLOR_INCLUDED

#include <stdio.h>
#include <float.h>

class Color {
	public :
		Color(float red, float green, float blue, float alpha);
		Color();
		float r, g, b, a;
		
		unsigned long composite;		// 32-bit composite color
		
		Color copy();	// make a copy of this color
		
		void print();
		
		// after adjusting the r, g, b, a values exterally
		// call this to remake the composite unsigned long
		unsigned long rebuildComposite();
		
		// get largest component of R,G,B
		float getMax();
		
		// alter color data by multiplying by weight
		void weightColor( float weight );
		
		// get component by component weighted 32-bit composite
		// (returns alpha unweighted)
		unsigned long getWeightedComposite( float weight );		// from this color
		unsigned long getWeightedComposite(unsigned long c1, float weight );	// from composite
		
		unsigned long sumComposite(unsigned long c1, unsigned long c2);
		
		// access this color as a three vector
		float &operator[](int rgbIndex);	
	};
	
	
inline Color::Color() {
	r = 0;
	g = 0;
	b = 0;
	a = 0;
	
	composite = 0;
	}

inline Color::Color(float red, float green, float blue, float alpha=0) {
	r = red;
	g = green;
	b = blue;
	a = alpha;
	
	composite = ((int)(b*255)) | ((int)(g*255)) << 8 | ((int)(r*255)) << 16 | ((int)(a*255)) << 24;
	}

inline Color Color::copy() {
	Color copyColor(r,g,b,a);
	return copyColor;
	}

inline void Color::print() {
	printf( "(%f, %f, %f)", r, g, b );
	}

inline unsigned long Color::rebuildComposite() {
	composite = ((int)(b*255)) | ((int)(g*255)) << 8 | ((int)(r*255)) << 16 | ((int)(a*255)) << 24;
	return composite;
	}

inline float Color::getMax() {
	float max = -FLT_MAX;
	if( r > max ) max = r;
	if( g > max ) max = g;
	if( b > max ) max = b;
	return max;
	}
	

inline void Color::weightColor( float weight ) {
	r = r * weight;
	g = g * weight;
	b = b * weight;
	// for now, don't touch alpha
	composite = ((int)(b*255)) | ((int)(g*255)) << 8 | ((int)(r*255)) << 16 | ((int)(a*255)) << 24;
	}	

inline float &Color::operator[](int rgbIndex) {
	if( rgbIndex == 0) return r;
	else if( rgbIndex == 1) return g;
	else if( rgbIndex == 2) return b;
	else if( rgbIndex == 3) return a;
	return r;		// default, return r reference
	}
	
	
inline unsigned long Color::getWeightedComposite( float weight ) {
	return ((int)(b*255*weight)) | ((int)(g*255*weight)) << 8 | ((int)(r*255*weight)) << 16 | ((int)(a*255)) << 24;
	}


inline unsigned long Color::getWeightedComposite( unsigned long c1, float weight ) {
	int b = c1 & 0xFF;
	
	int g = (c1 >> 8) & 0xFF;
	
	int r = (c1 >> 16) & 0xFF;

	int a = c1 >> 24;
	
	return ((int)(b*weight)) | (((int)(g*weight)) << 8) | (((int)(r*weight)) << 16) | (((int)(a*weight)) << 24);
	}

inline unsigned long Color::sumComposite(unsigned long c1, unsigned long c2) {
	int b = (c1 & 0xFF) + (c2 & 0xFF);
	if( b > 255) b = 255;
	
	int g = ((c1 >> 8) & 0xFF) + ((c2 >> 8) & 0xFF);
	if( g > 255) g = 255;
	
	int r = ((c1 >> 16) & 0xFF) + ((c2 >> 16) & 0xFF);
	if( r > 255) r = 255;

	int a = (c1 >> 24) + (c2 >> 24);
	if( a > 255) a = 255;

	return b | (g << 8) | (r << 16) | (a << 24);
	}
#endif