// Jason Rohrer
// Output.cpp

/**
*
*	Implementation for outputing textures computed by RadiosGL
*
*
*	Created 11-14-99
*	Mods:
*		Jason Rohrer	12-4-99		Added Delaunay triangulation for sub-patch centers interpolation
*		Jason Rohrer	12-5-99		Added adaptive patch texture sizing. (Decides based on patch area)
*		Jason Rohrer	12-16-99	When making a texture for a patch, all coplanar pathches that share
*										edges with the patch are also projected
*
*		Jason Rohrer	1-16-2000	Changed to use UniversalFileIO to work on both big 
*									and little endian machines
*
*		Jason Rohrer	1-17-2000	Changed so that texture anchors don't but right against texture edge
*									This was causing problems with including neighboring patches
*									as influencing this patch's texture, since neighbors were being cut off
*
*		Jason Rohrer	1-30-2000	Fixed bug in outputting texture sizes
*/



#include "Output.h"
#include "UniversalFileIO.h"


// prototypes for local functions
void interpolateLine( unsigned long *buffer, int *yOffset, int x1, int y1, int x2, int y2, Color *c1, Color *c2);
int log2(int x);	// integer power of 2 closest to x
int pow2( int pow );	// returns 2^pow

Output::Output( FILE *oFile, Patch *patchesIn, int nPatches, int maxTxHigh, int maxTxWide, float maxPArea, SimpleVector<int> *shrEdges ) {

	UniversalFileIO fileIO;

	outFile = oFile;
	patches = patchesIn;
	numPatches = nPatches;
	maxTextHigh = maxTxHigh;
	maxTextWide = maxTxWide;
	
	maxPatchArea = maxPArea;
	
	shareEdges = shrEdges;
	
	// set to 0 for testing
	float percentFlange = 0.10;		// added 1-17-2000:  gutter amount between texture anchors and texture edge
	
	
	// write number of patches into file
	//fwrite( (void *)(&numPatches), sizeof( long ), 1, outFile );
	fileIO.fwriteLong( outFile, numPatches );
	
	
	// setup texture sizes
	texturesWide = new long[ nPatches ];
	texturesHigh = new long[ nPatches ];
	
	// setup rotation angle arrays
	thetaX = new float[ nPatches ];
	thetaY = new float[ nPatches ];
	
	// setup zoom arrays
	zoomX = new float[ nPatches ];
	zoomY = new float[ nPatches ];
	
	// setup shift arrays
	shiftX = new float[ nPatches ];
	shiftY = new float[ nPatches ];
	
	// setup anchor arrays
	for( int i=0; i<3; i++ ) {
		textAnchorX[i] = new long[ nPatches ];
		textAnchorY[i] = new long[ nPatches ];
		}
	
	float invMaxPatchAreaRoot = 1.0 / sqrt(maxPatchArea);
	
	// fill in arrays
	for( int p=0; p<numPatches; p++ ) {		// for each patch
		
		// calculate rotation needed to bring patch into XY plane
		Vertex norm = patches[p].getNormal();
		norm.calcRotationToZ( thetaX[p], thetaY[p] );
		
		
		// calculate texture size based on area
		float area = patches[p].getArea();
		if( area < maxPatchArea ) {
			float fractionOfLargest = sqrt(area) * invMaxPatchAreaRoot;
			
			float thisTextWide = maxTextWide * fractionOfLargest;
			float thisTextHigh = maxTextHigh * fractionOfLargest;
			
			// find closest power of two to these sizes
			int powWide = log2( (int) thisTextWide );
			int powHigh = log2( (int) thisTextHigh );
			
			texturesWide[p] = pow2( powWide );
			texturesHigh[p] = pow2( powHigh );
			
			// make sure textures are big enough
			if( texturesHigh[p] < 4 ) { 
				texturesHigh[p] = 4;
				}
			if( texturesWide[p] < 4 ) { 
				texturesWide[p] = 4;
				}
			}
		else {
			texturesWide[p] = maxTextWide;
			texturesHigh[p] = maxTextHigh;
			}
		
		long textHigh = texturesHigh[p];	// take out of arrays for local use
		long textWide = texturesWide[p];
		
		// make a copy of this patch and rotate it into XY plane
		Patch *pCopy = patches[p].copy(); 
		pCopy->rotateX( thetaX[p] );
		pCopy->rotateY( thetaY[p] );
		
		
		// find max width of patch
		float maxX = - FLT_MAX;
		float minX = FLT_MAX;
		
		for( int v=0; v<3; v++ ) {
			if( pCopy->vertices[v].x > maxX ) {
				maxX = pCopy->vertices[v].x;
				}
			if( pCopy->vertices[v].x < minX ) {
				minX = pCopy->vertices[v].x;
				}
		
			}
		
		float widthX = maxX - minX;
		
		float flange = textWide * percentFlange;
		
		zoomX[p] = (textWide - 2*flange) / widthX;
		
		float zoomedMinX = zoomX[p] * minX;
		
		shiftX[p] = -zoomedMinX + flange;
		
		
		
		// find max height of patch
		float maxY = - FLT_MAX;
		float minY = FLT_MAX;
		
		for( int v=0; v<3; v++ ) {
			if( pCopy->vertices[v].y > maxY ) {
				maxY = pCopy->vertices[v].y;
				}
			if( pCopy->vertices[v].y < minY ) {
				minY = pCopy->vertices[v].y;
				}
		
			}
		
		float heightY = maxY - minY;
		
		flange = textHigh * percentFlange;
		
		zoomY[p] = (textHigh - 2*flange) / heightY;
		
		float zoomedMinY = zoomY[p] * minY;
		
		shiftY[p] = -zoomedMinY + flange;
		
		
		// now that zoom and shift done, use to calculate texture anchors
		for( int v=0; v<3; v++ ) {		// for each vertex in patch
			int x = (int)( pCopy->vertices[v].x * zoomX[p] + shiftX[p] );
			int y = (int)( pCopy->vertices[v].y * zoomY[p] + shiftY[p] );
			if( x < 0 )  x = 0;
			if( x >= textWide ) x = textWide - 1;
			if( y < 0 )  y = 0;
			if( y >= textHigh ) y = textHigh - 1; 
			textAnchorX[v][p] = x;
			textAnchorY[v][p] = y;
			}
			
		delete pCopy;
		
		}
	}
		
	
	
			
		
void Output::buildFromGroups( SimpleVector<PatchGroup*> &groups ) {
	
	UniversalFileIO fileIO;
	
	int numGroups = groups.size();


	printf("writing texture data for patch:  ");


	// for each patch, find sub patches that contribute
	
	for( int p=0; p<numPatches; p++ ) {		// for each patch
		
		printf("%d/%d   ", p, numPatches);
		
		long textHigh = texturesHigh[p];	// take out of arrays for local use
		long textWide = texturesWide[p];
		
		unsigned long *texture = new unsigned long[ textHigh * textWide ];
		// precalculate y offsets into texture
		int *yOffset = new int[ textHigh ];
		for( int y=0; y<textHigh; y++ ) {
			yOffset[y] = y * textWide;
			}
		
		// empty the texture for this patch
		for( int y=0; y<textHigh; y++ ) {
			int yContrib = yOffset[y];
			for( int x=0; x<textWide; x++ ) {
				// alpha of 0 means there is no patch here yet
				texture[ yContrib + x ] = 0;
				}
			}
		
		int numGroups = groups.size();
		
		int numSubPatches = 0;
		// first, count number of subgroups for this patch
		for( int g=0; g<numGroups; g++ ) {		// for each group
			
			int parID = (*(groups.getElement(g)))->getParentPatchID();
			
			// search for this ID inside shared edge vector to see if it shares edges with this patch
			char groupSharesEdges = false;
			int numShare = shareEdges[p].size();
			for( int q=0; q<numShare; q++ ) {
				groupSharesEdges = groupSharesEdges || ( *(shareEdges[p].getElement( q )) == parID);
				}

			if( parID == p ) {		// this group part of this patch
				// four patches in this group
				numSubPatches += 4;
				}
			if( groupSharesEdges ) {		// this group part of neighboring patch
				// four patches in this group
				numSubPatches += 4;
				}
			}
		
		if( numSubPatches < 4 ) {
			printf("debug\n");
			}	
		// create structures of numSubPatches size to store points for triangulation
		struct triangulateio in, out;
		// input
		in.numberofpoints = numSubPatches;
		in.numberofpointattributes = 0;
		in.pointlist = new REAL[ numSubPatches * 2];		// 2 REALs per X,Y point
		in.pointmarkerlist = NULL;
		in.numberofsegments = 0;
		in.numberofholes = 0;
		in.numberofregions = 0;
		
		// output:  let Triangle do the memory allocation
		out.edgelist = (int *) NULL;	
		// only set edgelist to null, since switches tell Triangle only to produce an edge list
			
		
		// Call to triangle:  (called later)
		// N,E,P,B	-- suppress the output node, element, poly, and marker info
		// Q		-- quiet mode
		// e -- output edge information
		//triangulate("pczNEPBQe", &in, &out, (struct triangulateio *)NULL);
	
		
		Color **centerColors = new Color*[numSubPatches];	// pointers to colors of each center
		
		
		// now add points to triangulation structure
		int currentPoint = 0;		
		
		for( int g=0; g<numGroups; g++ ) {		// for each group
			
			int parID = (*(groups.getElement(g)))->getParentPatchID();
			
			// search for this ID inside shared edge vector to see if it shares edges with this patch
			char groupSharesEdges = false;
			int numShare = shareEdges[p].size();
			for( int q=0; q<numShare; q++ ) {
				groupSharesEdges = groupSharesEdges || (*(shareEdges[p].getElement( q )) == parID);
				}
			
			// check if this group is part of this patch (or if group part of a neigboring coplanar patch)
			if( parID == p || groupSharesEdges ) {
				
				// will use these to fill in interpolated edges between each pair of patch centers
				int fourCenters[4][2];	// x and y coords of 4 patch centers
				Color *fourColors[4];		// color of each patch center
				
				for( int s=0; s<4; s++ ) {		// for each patch in group
				
					Patch *subPatch = (*(groups.getElement(g)))->getPatch(s);
					
					// rotate a copy of this patch so that it is lined up with xy plane
					Patch *pCopy = subPatch->copy();
					pCopy->rotateX( thetaX[p] );
					pCopy->rotateY( thetaY[p] );
					
					Vertex center = pCopy->getCenter();
					
					int x = (int)( center.x * zoomX[p] + shiftX[p] );
					int y = (int)( center.y * zoomY[p] + shiftY[p] ); 
					
					
					// can no longer assume center is inside texture:
					//		we're projecting groups that aren't from this patch now (neighboring groups)
	
					// handle this problem before interpolating triangulated points
					
	
					// add this center to list
					fourCenters[s][0] = x;
					fourCenters[s][1] = y;
					fourColors[s] = (*(groups.getElement(g)))->getRadiosity(s);
					
					//add center point of sub patch to texture
			// This doesn't work now that we may have centers outside texture border
			// no need to do this anyway
			//		texture[ yOffset[y] + x ] = ((*(groups.getElement(g)))->getRadiosity(s))->composite;
					
					delete pCopy;
					}
				
				
				// now have 4 centers
				// add these to running triangulation
				for( int m=0; m<4; m++ ) {
					int x = fourCenters[m][0];
					int y = fourCenters[m][1];
					
					in.pointlist[currentPoint] = (REAL)x;
					in.pointlist[currentPoint + 1] = (REAL)y;
					
					// currentPoint / 2 since there's only one color per point (not two coords like pointlist)
					centerColors[currentPoint >> 1] = fourColors[m];		// save pointer to color
					
					currentPoint += 2;
					}
				}	// end check if this group part of patch p
			}	// end for each group
			
	// have added all center points to texture and triangulation structures
	// now run triangulation
	triangulate("pczNEPBQe", &in, &out,(struct triangulateio *)NULL);		// switches explained above
	
	// triangulation done... now walk through edges and interpolate across them
	int numEdges = out.numberofedges;
	int numEndpoints = numEdges << 1;	// two points per edge
	for( int v=0; v<numEndpoints; v+=2 ) {
		int startInd = out.edgelist[v];		// index into points of edge start
		int endInd = out.edgelist[v+1];		// edge end
		
		// colors match 1-1 point indexing
		int colorStartInd = startInd;	
		int colorEndInd = endInd;
		
		// vertices have 2 pointlist vals each
		startInd = startInd << 1;
		endInd = endInd << 1;
		
		// index into pointlist to get coords of start and end points
		int startX = (int) ( in.pointlist[startInd] );
		int startY = (int) ( in.pointlist[startInd + 1] );
		int endX = (int) ( in.pointlist[endInd] );
		int endY = (int) ( in.pointlist[endInd + 1] );
		
		// make sure start and end points are in bounds
		if( startX >= 0 && startX < textWide ) {
			if( startY >= 0 && startY < textHigh) {
				if( endX >= 0 && endX < textWide ) {
					if( endY >= 0 && endY < textHigh) {
						interpolateLine( texture, yOffset, startX, startY, endX, endY, centerColors[colorStartInd], centerColors[colorEndInd]);
						}
					}
				}
			}
		}		
	
	delete [] in.pointlist;
	delete [] out.edgelist;
	delete [] centerColors;
	
	// now have triangulation of all patch centers interpolated in texture
	// use bilinear interpolation to fill in the rest of the pixels
		
		
		// now have one pixel lines in texture for each sub patch...
		// need to fill in rest using linear interpolation
		
		// first, interpolate horizontally
		
		Color dummyC;
		
		for( int y=0; y<textHigh; y++ ) {		// for each row
			int yContrib = yOffset[y]; 
			
			int currentX = 0;
			int startRegion = -1;
			int endRegion = -1;
			
			// first, find first region start
			for( currentX=0; currentX<textWide; currentX++ ) {
				if( (texture[ yContrib + currentX ] >> 24) > 0 ) {		// if alpha not 0
					startRegion = currentX;
					currentX = textWide;	// found, so jump out
					}
				}
			if( startRegion != -1 ) {		// found a colored region
				endRegion = startRegion;
				startRegion = 0;
				currentX = startRegion;
				
				unsigned long endColor = texture[ yContrib + endRegion ];
				unsigned long startColor = endColor;
				
				while( currentX < textWide-1 ) {
					// walk through current region and fill in colors
					
					float invRegionLength = 1.0 / ( endRegion - startRegion );
					
					for( currentX=startRegion; currentX<=endRegion; currentX++ ) {
						float endWeight = (currentX - startRegion) * invRegionLength;
						float startWeight = (endRegion - currentX) * invRegionLength;
					
						// FIXME:  Rounding errors here may cause problems
						//		(converting to and from ints with float weights is problematic)
					
						unsigned long weightedStartColor = dummyC.getWeightedComposite( startColor, startWeight );
						unsigned long weightedEndColor = dummyC.getWeightedComposite( endColor, endWeight );
						texture[ yContrib + currentX ] = dummyC.sumComposite( weightedStartColor, weightedEndColor );
						}
					
					// new region start == old region end
					startRegion = endRegion;
					endRegion = textWide-1;		// if none found
					for( currentX=startRegion+1; currentX<textWide; currentX++ ) {		// look for next end region
						if( (texture[ yContrib + currentX ] >> 24) > 0 ) {		// if alpha not 0
							endRegion = currentX;
							currentX = textWide;	// found, so jump out
							}
						}	
					
					startColor = endColor;
					if( (texture[ yContrib + endRegion] >> 24) > 0 ) {		// if alpha not 0 of end region
						endColor = texture[ yContrib + endRegion];
						}
					else {
						endColor = startColor;
						}
					currentX = startRegion;		// to jump out of loop once done
					
					
					}	// end while currentX < textWide -1
					
				}	// end check for colored region in this row
			
			}	// end for each row
		
		
		// now all fillable rows have been filled in....
		// do the same thing for each column
		for( int x=0; x<textWide; x++ ) {		// for each column
			
			int currentY = 0;
			int startRegion = -1;
			int endRegion = -1;
			
			// first, find first region start
			for( currentY=0; currentY<textHigh; currentY++ ) {
				if( (texture[ yOffset[currentY] + x ] >> 24) > 0 ) {		// if alpha not 0
					startRegion = currentY;
					currentY = textHigh;	// found, so jump out
					}
				}
			// guarenteed to be a colored region in this column....
			endRegion = startRegion;
			startRegion = 0;
			currentY = startRegion;
			
			unsigned long endColor = texture[ yOffset[endRegion] + x ];
			unsigned long startColor = endColor;
			
			while( currentY < textHigh-1 ) {
				// walk through current region and fill in colors
				
				float invRegionLength = 1.0 / ( endRegion - startRegion );
				
				for( currentY=startRegion; currentY<=endRegion; currentY++ ) {
					float endWeight = (currentY - startRegion) * invRegionLength;
					float startWeight = (endRegion - currentY) * invRegionLength;
					
					// FIXME:  Rounding errors here may cause problems
					//		(converting to and from ints with float weights is problematic)
					
					unsigned long weightedStartColor = dummyC.getWeightedComposite( startColor, startWeight );
					unsigned long weightedEndColor = dummyC.getWeightedComposite( endColor, endWeight );
					texture[ yOffset[currentY] + x ] = dummyC.sumComposite( weightedStartColor, weightedEndColor );
					}
				
				// new region start == old region end
				startRegion = endRegion;
				endRegion = textWide-1;		// if none found
				for( currentY=startRegion+1; currentY<textHigh; currentY++ ) {		// look for next end region
					if( (texture[ yOffset[currentY] + x ] >> 24) > 0 ) {		// if alpha not 0
						endRegion = currentY;
						currentY = textHigh;	// found, so jump out
						}
					}	
				
				startColor = endColor;
				if( (texture[ yOffset[endRegion] + x ] >> 24) > 0 ) {		// if alpha not 0 of end region
					endColor = texture[ yOffset[endRegion] + x];
					}
				else {
					endColor = startColor;
					}
				currentY = startRegion;		// to jump out of loop once done
				
				
				}	// end while currentY < textHigh -1
				
			
			}	// end for each column
	
		
		// texture should be filled in here
		
		int postSmooth = true;
		int smoothRadius = 10;
		
		if( postSmooth ) {
			// smooth the texture
			smoothTexture( textHigh, textWide, texture, smoothRadius );
			}

		// write out vertices
		for( int v=0; v<3; v++ ) {
			float x = patches[p].vertices[v].x;
			float y = patches[p].vertices[v].y;
			float z = patches[p].vertices[v].z;
			
			fileIO.fwriteFloat( outFile, x );
			fileIO.fwriteFloat( outFile, y );
			fileIO.fwriteFloat( outFile, z );
			}
		
		// write out texture information
		fileIO.fwriteLong( outFile, textWide );
		fileIO.fwriteLong( outFile, textHigh );
		
		// write texture anchors
		for( int v=0; v<3; v++ ) {
			long x = textAnchorX[v][p];
			long y = textAnchorY[v][p];
			fileIO.fwriteLong( outFile, x );
			fileIO.fwriteLong( outFile, y );
			}
			
		// write texture data:
		//fwrite( (void *)texture, sizeof(long), textHigh*textWide, outFile);				
		
		// NOTE:  this new way (using UniversalFileIO) is MUCH more inefficient
		// This may not matter much in the long run, though, especially since
		//   file IO is the least of our time worries.
		long numPixels = textHigh * textWide;
		for( int pixel=0; pixel<numPixels; pixel++ ) {
			fileIO.fwriteLong( outFile, texture[pixel] );
			}
			
		// free memory associated with this texture
		delete [] texture;
		delete [] yOffset;
	
		}	// end for all patches

	printf("\nDone writing texture data.\n");

	}
	
	
// interpolate colors along a line in a buffer	
// assumes line is in the buffer (no bounds checking
void interpolateLine( unsigned long *buffer, int *yOffset, int x1, int y1, int x2, int y2, Color *c1, Color *c2) {
	
	int startY = y1;
	int endY = y2;
	int startX = x1;
	int endX = x2;
	
	Color *startColor = c1;
	Color *endColor = c2;
	
	if( y1 > y2 ) {		// make sure y increases from start to end
		startY = y2;
		endY = y1;
		startX = x2;
		endX = x1;
		startColor = c2;
		endColor = c1;
		}
	
	float slopeX = (float)(endX - startX) / (float)(endY - startY);
	
	float currentWeight = 0;
	float slopeWeight = 1.0 / (float)(endY - startY);
	
	float currentX = (float)startX;
	
	for( int y=startY; y<=endY; y++ ) {
		int yContrib = yOffset[y];
		
		// weight each color
		unsigned long compStartWeighted = startColor->getWeightedComposite( 1.0 - currentWeight );	
		unsigned long compEndWeighted = endColor->getWeightedComposite( currentWeight );
		
		// sum the weighted composites
		unsigned long colorSum =  startColor->sumComposite( compStartWeighted, compEndWeighted );
		
		
		// set the buffer to this color
		buffer[ yContrib + (int)currentX ] = colorSum;
		
		currentX += slopeX;
		currentWeight += slopeWeight;
		}	
	}	
	
	
int log2( int x ) {
	int numDivides = 0;
	
	while( x != 0 ) {
	
		x = x >> 1;
		numDivides++;
		}
	
	return numDivides;
	}
	
int pow2( int pow ) {
	return 1 << pow;
	}