// Jason Rohrer
// Patch.h

/**
*
*	Patch object for RadiosGL
*
*	3 vertices
*	1 normal
*	1 color
*
*
*	Created 9-5-99
*	Mods:
*
*		Jason Rohrer	12-1-99		Added patch area function
*									Optimization:
*									Improved performance by calculating sine and cosine outside of 
*									each vertex rotation
*									Optimization:
*									Precompute and rotate normal with patch
*										(see comment below for explaination)
*									Optimization:
*									The last optimizaiton didn't work, so it was removed
*
*		Jason Rohrer	12-16-99	Added functions for testing coplanar and shared edge properties
*
*		Jason Rohrer	1-16-2000	Changed shareEdges function so it catches edges that are
*									partially shared
*
*		Jason Rohrer	1-17-2000	Changed shareEdges function so it catches any patches that
*									share a vertex.  This should correct many of the shading problems
*
*		Jason Rohrer	1-18-2000	Changed shareEdges back to the way it was... catching shared
*									vertices alone didn't fix the shading problems
*
*		Jason Rohrer	1-30-2000	Added print function
*/


#ifndef PATCH_INCLUDED
#define PATCH_INCLUDED

#include "Vertex.h"
#include "RadiosityConstants.h"


#include <math.h>

class Patch {

	public :
	
		Patch(Vertex a, Vertex b, Vertex c);
		Patch();
		void setupPatch(Vertex a, Vertex b, Vertex c);	// setup patch after construction
		
		
		~Patch();
		
		Vertex *vertices;
		
		// Vertex norm;
		
		void move(Vertex &deltaVert);	// move by delta distance
		
		void rotateX(float thetaX);		// rotate around x axis
		void rotateY(float thetaY);		// rotate around y axis
		void rotateZ(float thetaZ);		// rotate around z axis

		Patch *copy();		// make a copy of this patch
		
		Vertex getCenter();
		Vertex getNormal();
		
		float getArea();	// get area of this patch
		
		char areCoplanar( Patch *other );		// returns true if these are coplanar
												// also returns true if they lie in || planes
		
		char shareEdges( Patch *other );		// returns true if these patches share and edge
		char shareVertex( Patch *other );		// returns true if these patches share and edge (two vertices)
		
		void print();
		
	private :
		
		float tolerance;		// error to tolerate when testing for shared edges

	};	
		






inline Patch::Patch() {
	vertices = new Vertex[3];
	tolerance = 2;
	}

inline void Patch::setupPatch(Vertex a, Vertex b, Vertex c) {
	vertices[0] = a;
	vertices[1] = b;
	vertices[2] = c;
	}

inline Patch::Patch( Vertex a, Vertex b, Vertex c) {
	
	vertices = new Vertex[3];
	
	vertices[0] = a;
	vertices[1] = b;
	vertices[2] = c;
	}
	


inline Patch::~Patch() {
	delete [] vertices;
	}
	

inline void Patch::move(Vertex &deltaVert) {
	for(int i=0; i<3; i++) {	// don't use vector overloads here to prevent memory leaks
		vertices[i].move(deltaVert);
		}
	} 

// NOTE:
//     Doesn't seem that normals are really needed for the whole
//		radiosity process...  Form factors seem to handle this....
//		so for now, don't rotate normals, but add later if needed

// Optimization:  precompute and rotate normal with patch
//		After every rotation, normal is used to see if patch faces zbuffer
//		Recalculating normal each time requires a sqrt function call
//		Rotating throughout is faster, since sin and cos need to be computed for vertex rotation anyway

// This optimization didn't work.... getNormal not used often enough to make it worth it

inline void Patch::rotateX(float thetaX) {
	float sinTheta = sin( thetaX );
	float cosTheta = cos( thetaX );
	
	for( int i=0; i<3; i++) {
		vertices[i].rotateX( sinTheta, cosTheta );
		}
	}
	
inline void Patch::rotateY(float thetaY) {
	float sinTheta = sin( thetaY );
	float cosTheta = cos( thetaY );
	
	for( int i=0; i<3; i++) {
		vertices[i].rotateY( sinTheta, cosTheta );
		}
	}
		
inline void Patch::rotateZ(float thetaZ) {
	float sinTheta = sin( thetaZ );
	float cosTheta = cos( thetaZ );
	
	for( int i=0; i<3; i++) {
		vertices[i].rotateZ( sinTheta, cosTheta );
		}
	}

inline Patch *Patch::copy() {
	Patch *returnPatch = new Patch( vertices[0].copy(),  vertices[1].copy(),  vertices[2].copy() );
	return returnPatch;
	}

inline Vertex Patch::getCenter() {
	Vertex returnVert = (vertices[0] + vertices[1] + vertices[2]) * oneThird;
	return returnVert;
	}
	

inline Vertex Patch::getNormal() {
	Vertex v = vertices[2] - vertices[0];
	Vertex w = vertices[1] - vertices[0];
	
	Vertex normal(v.y * w.z - v.z * w.y, v.z * w.x - v.x * w.z, v.x * w.y - v.y * w.x);
	
	float length = sqrt( normal.x * normal.x + normal.y * normal.y + normal.z * normal.z );
	normal = normal / length;
	
	return normal;
	}



inline float Patch::getArea() {
	Vertex v = vertices[2] - vertices[0];
	Vertex w = vertices[1] - vertices[0];
	
	Vertex normal(v.y * w.z - v.z * w.y, v.z * w.x - v.x * w.z, v.x * w.y - v.y * w.x);
	
	float length = sqrt( normal.x * normal.x + normal.y * normal.y + normal.z * normal.z );
	
	// length of normal is area of parallelogram formed by these edges
	// area of triangle is half of this area
	return length * 0.5;
	}


inline char Patch::areCoplanar( Patch *other ) {
	// simple method:
	// normals for each patch:  their dot should be 1.0
	
	Vertex thisNormal = getNormal();
	Vertex otherNormal = other->getNormal();
	
	if( dot( thisNormal, otherNormal ) > 0.99 ) {		// almost 1, to allow for float errors
		return true;
		}
		
	return false;
	}


inline char Patch::shareVertex( Patch *other ) {		// returns true if patches share a vertex
	for( int m=0; m<3; m++ ) {
		Vertex otherVert = other->vertices[m];
		
		
		for( int n=0; n<3; n++ ) {
			Vertex thisVert = vertices[n];
			if( otherVert.x < thisVert.x + tolerance && otherVert.x > thisVert.x - tolerance ) {
				if( otherVert.y < thisVert.y + tolerance && otherVert.y > thisVert.y - tolerance ) { 
					if( otherVert.z < thisVert.z + tolerance && otherVert.z > thisVert.z - tolerance ) {
						return true;		// one shared vertex found
						}
					}
				}
			}
		}
	return false;		// no shared vertices found
	}
	

			
inline char Patch::shareEdges( Patch *other ) {		// returns true if these patches share and edge (two vertices)	
	
	// if they have any 2 verts in common, they share an edge
	
	char foundOne = false;
	char foundTwo = false;
	
	for( int m=0; m<3; m++ ) {
		Vertex otherVert = other->vertices[m];
		
		
		for( int n=0; n<3; n++ ) {
			Vertex thisVert = vertices[n];
			
			if( otherVert.x == thisVert.x && otherVert.y == thisVert.y && otherVert.z == thisVert.z ) {
				if( foundOne ) {
					foundTwo = true;
					}
				foundOne = true;
				}
			}
		}
	
	return foundOne && foundTwo;
	}


inline void Patch::print() {
	for( int m=0; m<3; m++ ) {
		printf("   Vertex %d: ", m);
		vertices[m].print();
		}
	}
	
	
#endif