// Jason Rohrer
// PatchGroup.cpp

/**
*
*	Group of three patches for RadiosGL
*
*	This group contains 3 patches that were orignially pieces of the same
*	patch.  Radiosities are calculated for these, and then the 3 patches
*	are subdivided further as needed.
*
*
*	Created 10-2-99
*	Mods:
*		Jason Rohrer	10-30-99	Changed subdivison (and group itself) to use fourths
*									instead of thirds.
*		
*		Jason Rohrer 	11-20-99	Fixed so that emittance not divided during patch division
*									( emittance is a per-unit-area value )
*
*		Jason Rohrer	1-30-2000	Fixed subdivision code.
*									Used to declare an array pointer to 4 new patches...
*									That caused a segmentation fault on little-endian machines
*									Now calls "new" four times to fill in an array of pointers.
*									There still may be an outstanding bug here...
*/


#include "PatchGroup.h"


void PatchGroup::setupPatches(Patch *p, int ID, Color *refl, Color *emitted, Color *radios, Hemicube *hemCube) {
	
	parentPatchID = ID;
	
	
	// take average of vertices to get midpoint of patch edge
	//Vertex midpoint = (p->vertices[0] + p->vertices[1] + p->vertices[2]) * oneThird; 
	Vertex midSide0 = (p->vertices[0] + p->vertices[1]) *.5;
	Vertex midSide1 = (p->vertices[1] + p->vertices[2]) *.5;
	Vertex midSide2 = (p->vertices[2] + p->vertices[0]) *.5;
	
	
	// four patches:
	// 		one with	v0, mid0, mid2
	//					v1, mid1, mid0
	//					v2, mid2, mid1
	//					mid0, mid1, mid2
	
	
	for( int i=0; i<4; i++) {
		fourPatches[i] = new Patch();
		}
	
	fourPatches[0]->vertices[0] = p->vertices[0].copy();
	fourPatches[0]->vertices[1] = midSide0.copy();
	fourPatches[0]->vertices[2] = midSide2.copy();
	
	fourPatches[1]->vertices[0] = p->vertices[1].copy();
	fourPatches[1]->vertices[1] = midSide1.copy();
	fourPatches[1]->vertices[2] = midSide0.copy();
	
	fourPatches[2]->vertices[0] = p->vertices[2].copy();
	fourPatches[2]->vertices[1] = midSide2.copy();
	fourPatches[2]->vertices[2] = midSide1.copy();
	
	fourPatches[3]->vertices[0] = midSide0.copy();
	fourPatches[3]->vertices[1] = midSide1.copy();
	fourPatches[3]->vertices[2] = midSide2.copy();
	
	reflectance = refl;
	
	for(int i=0; i<4; i++) {
		fourEmittances[i] = emitted;		// just keep pointer to emitted
		fourRadiosities[i] = new Color(radios->r, radios->g, radios->b, 1.0);
		}
	
	hCube = hemCube;
	
	needsSubDivide = false;
	
	beenSetup = true;
	}
	
	
void PatchGroup::project( SimpleVector<PatchGroup*> &groups ) {
	
	int numPatches = (groups.size())*4;
	for(int f=0; f<4; f++) {
		fourFormFactors[f] = new float[numPatches];
		for(int p=0; p<numPatches; p++) {	// set for factors to 0
			fourFormFactors[f][p] = 0.0;
			}
		}
	formFactorArraysCreated = true;
	
	numGroups = groups.size();
	
	for(int p=0; p<4; p++) {		// for each patch in this group
		
		Vertex center = fourPatches[p]->getCenter();
		Vertex normal = fourPatches[p]->getNormal();
		
		// create viewcube for this patch
		ViewCube *vCube = new ViewCube(center, normal, hCube);
			
		for( int g=0; g<numGroups; g++) {	// for each other group
			PatchGroup *pg = *(groups.getElement(g));
			int fourG = 4*g;	// precalc
			
			if(pg->getParentPatchID() != parentPatchID) {		// if this other group not in same subgroup... 
				
				for(int i=0; i<4; i++) {	// for each patch in other group
					Patch *otherPatch = pg->getPatch(i);
					
					int tempPatchID = (fourG) + i;		// create unique patch ID for this round
					vCube->projPatchOnCube( tempPatchID, otherPatch);		// project patch onto viewcube
					}
				}
			}		

		// now sum form factors for this patches view cube
		vCube->sumFormFactors( fourFormFactors[p] );

		delete vCube;		// form factors have been computed, view cube no longer needed
		}	
	}
	
	
	
	
char PatchGroup::subdivide( SimpleVector<PatchGroup*> &groups, int thisIndex ) {
	if( needsSubDivide ) {

		PatchGroup *newGroups[4];

		for(int i=0; i<4; i++) {
			newGroups[i] = new PatchGroup();
			newGroups[i]->setupPatches( fourPatches[i], parentPatchID, reflectance, fourEmittances[i], fourRadiosities[i], hCube );
			}
		
		// remove old group from vector
		groups.deleteElement( thisIndex );
		
		// add pointers to new groups to vector
		for(int i=0; i<4; i++) {
			groups.push_back( newGroups[i] );
			}	
		return true;	
		}
	return false;
	}


void PatchGroup::checkThreshold( float threshold ) {
	// for each pair of patches
	for( int m=0; m<4; m++ ) {
		for( int n=0; n<m; n++ ) {
			for( int c=0; c<3; c++ ) {	// for each color component
				// compare components between patches
				float difference = (*(fourRadiosities[m]))[c] - (*(fourRadiosities[n]))[c];
				if( difference > threshold || difference < -threshold ) {
					// threshold breached
					needsSubDivide = true;
					}
				}
			}
		}
	}	

	
void PatchGroup::clearRadiosity() {
	for(int i=0; i<4; i++) {
		for( int c=0; c<3; c++ ) {
			(*(fourRadiosities[i]))[c] = 0;
			}
		(*(fourRadiosities[i]))[3] = 1.0;		
		}	
	if( formFactorArraysCreated ) {
		for( int i=0; i<4; i++) {
			delete [] fourFormFactors[i]; // these arrays will change size in next round
			}
		}
	formFactorArraysCreated = false;	
	}

	
void PatchGroup::multFormFactorsByReflectance(int patchNum, int colorToMult) {
	for( int g=0; g<numGroups; g++) {	// for each other group	
		fourFormFactors[patchNum][g] = fourFormFactors[patchNum][g] * (*reflectance)[colorToMult];
		}
	}
		

void PatchGroup::divideFormFactorsByReflectance(int patchNum, int colorToDivide) {
	for( int g=0; g<numGroups; g++) {	// for each other group	
		fourFormFactors[patchNum][g] = fourFormFactors[patchNum][g] / (*reflectance)[colorToDivide];
		}
	}
	


void PatchGroup::print() {
	printf( "Parent patch ID = %d\n", parentPatchID );
	for( int g=0; g<4; g++) {	// for each other group	
		printf( "Patch %d = ", g );
		fourPatches[g]->print();
		printf( "\n" );
		}
	}


