// Jason Rohrer
// PatchGroup.h

/**
*
*	Group of three patches for RadiosGL
*
*	This group contains 3 patches that were orignially pieces of the same
*	patch.  Radiosities are calculated for these, and then the 3 patches
*	are subdivided further as needed.
*
*
*	Created 10-2-99
*	Mods:
*		Jason Rohrer	10-24-99	Fixes, changed vectors, fixed form factor calc, fixed destructor
*
*		Jason Rohrer	12-1-99		Added patch area function
*									Added force subdivide function
*/




#ifndef PATCHGROUP_INCLUDED
#define PATCHGROUP_INCLUDED

#include "SimpleVector.h"

#include "Hemicube.h"
#include "ViewCube.h"
#include "Patch.h"
#include "RadiosityConstants.h"

class PatchGroup {
	
	public :
		
		PatchGroup();
		
		// pass in a patch which will be subdivided to make up the group
		void setupPatches(Patch *p, int ID, Color *refl, Color *emitted, Color *radios, Hemicube *hemCube);
		
		~PatchGroup();

		// divide patch into four new groups (of four patches each)
		// remove ourselves from the vector	
		// return true if subdivide happened	
		char subdivide( SimpleVector<PatchGroup*> &groups, int thisIndex );
		
		// check if patch group needs subdivision based on sub-patch difference threshold
		// mark this group for division if necessary
		void checkThreshold( float threshold );
		
		// mark this patch for a manditory subdivide
		// (used during initial patch subdivision)
		void forceSubdivide();
		
		// reset all radiosities to 0
		// ( for after a restart )
		void clearRadiosity();
		
		
		float *getFormFactors(int patchNum);	// get ptr to array of form factors for this patch, 0..2
		
		Color *getEmittance(int patchNum);		// get ptr to emittance for this patch, 0..2
		
		Color *getRadiosity(int patchNum);		// get ptr to radiosity for patch, 0..2
		
		Color *getReflectance();	// get ptr to reflectance color
		
		
		float getPatchArea();	// returns area of a patch in this group (all should be the same)
		
		void multFormFactorsByReflectance(int patchNum, int colorToMult);		// premultiply for radiosity calc, 0..2 for rgb
		
		void divideFormFactorsByReflectance(int patchNum, int colorToDivide);		// return form factor to normal, 0..2 for rgb
		
		
		void project( SimpleVector<PatchGroup*> &groups );
		
		Patch *getPatch(int patchNum);
		
		int getParentPatchID();
		
		void print();		// print information about this patch group
		
	private:
		char beenSetup;
		char formFactorArraysCreated;
		
		int parentPatchID;
		
		
		Patch *fourPatches[4];
		Color *reflectance;
		Color *fourEmittances[4];
		Color *fourRadiosities[4];
		float *fourFormFactors[4];			// array of form factors with every other patch in scene
		
		int numGroups;
		
		Hemicube *hCube;
		
		char needsSubDivide;	// mark for future subdivision
	};


inline PatchGroup::PatchGroup() {
	beenSetup = false;
	formFactorArraysCreated = false;
	}


inline PatchGroup::~PatchGroup() {
	if( beenSetup ) {		// test if patch contains any allocated pointers
		for( int i=0; i<4; i++) {
			delete fourRadiosities[i];
			delete fourPatches[i];
			}
		if( formFactorArraysCreated ) {
			for( int i=0; i<4; i++) {
				delete [] fourFormFactors[i];
				}
			}
		}
	
	}



inline int PatchGroup::getParentPatchID() {
	return parentPatchID;
	}

inline Patch *PatchGroup::getPatch(int patchNum) {
	return fourPatches[patchNum];
	}

inline Color *PatchGroup::getRadiosity(int patchNum) {
	return fourRadiosities[patchNum];
	}
	
	
inline float *PatchGroup::getFormFactors(int patchNum) {
	return fourFormFactors[patchNum];
	}
	
inline Color *PatchGroup::getEmittance(int patchNum) {
	return fourEmittances[patchNum];
	}
	
inline Color *PatchGroup::getReflectance() {
	return reflectance;
	}

inline void PatchGroup::forceSubdivide() {
	needsSubDivide = true;
	}

inline float PatchGroup::getPatchArea() {
	return (fourPatches[0])->getArea();
	}
	
#endif
