// Jason Rohrer
// Radiosity.cpp

/**
*
*	Main radiosity object functions for RadiosGL
*
*
*	Created 10-2-99
*	Mods:
*		Jason Rohrer	10-24-99	Fleshed-out run function, finished constructor
*									put patch groups into a vector
*
*		Jason Rohrer	11-20-99	Added gamma correction to color normalization
*
*		Jason Rohrer	12-1-99		Added subdivision of patches before any computation
*									this is needed so that all patches are about the same size
*									Decide sub division based on patch area
*									User specified threshold for initial division
*
*		Jason Rohrer	12-16-99	Added check for coplanar patches that share edges
*
*		Jason Rohrer	1-24-2000	Added check for same reflectance and emittance
*									color before deciding that patches share an edge
*									This allows for a sharply divided checkerboard floor, if desired,
*									without being blended over.
*
*/

#include "Radiosity.h"


Radiosity::Radiosity(FILE *modlFile, FILE *hcFile, FILE *outFile, float hemicubeWorldWide, float maxPArea, float iterThresh, float subDivThresh, float gam, int textSize) {
	
	
	hemiC = new Hemicube( hemicubeWorldWide, hcFile);		// construct the hemicube
	iterThreshold = iterThresh;
	subDivideThreshold = subDivThresh;
	
	fscanf(modlFile, "%d", &numPatches);	// read number of patches from file
	
	patches = new Patch[numPatches];
	
	printf("num patches =%d\n", numPatches);
	
	
	// keep temp arrays for r, g, b and er, eg, eb colors 
	// to be used when deciding which patches are part of same planar object
	
	float *red = new float[numPatches];
	float *green = new float[numPatches];
	float *blue = new float[numPatches];
	
	float *ered = new float[numPatches];
	float *egreen = new float[numPatches];
	float *eblue = new float[numPatches];
	
	
	for( int p=0; p< numPatches; p++) {		// for each patch, read it from file
		printf("patch %d\n",p);
		float x,y,z;
		Vertex *verts[3];
		for( int v=0; v<3; v++) {		// for each vertex
			fscanf(modlFile, "%f", &x);
			fscanf(modlFile, "%f", &y);
			fscanf(modlFile, "%f", &z);
			
			
			printf("vertex read = (%f, %f, %f)\n", x, y, z);
			
			verts[v] = new Vertex(x, y, z);
			}
	
		patches[p].setupPatch(*(verts[0]), *(verts[1]), *(verts[2]));	// pass verts into patch
		
		// read color info for this patch
		
		float r, g, b;
		fscanf(modlFile, "%f", &r);
		fscanf(modlFile, "%f", &g);
		fscanf(modlFile, "%f", &b);
		
		red[p] = r;
		green[p] = g;
		blue[p] = b;
		
		// read emitted light info from file
		float er, eg, eb;
		fscanf(modlFile, "%f", &er);
		fscanf(modlFile, "%f", &eg);
		fscanf(modlFile, "%f", &eb);
		
		ered[p] = er;
		egreen[p] = eg;
		eblue[p] = eb;
		
		// now construct the patch group
		PatchGroup *pg = new PatchGroup();
		pg->setupPatches(&(patches[p]), p, new Color(r, g, b, 1.0), new Color(er,eg,eb, 1.0), new Color(0,0,0, 1.0), hemiC);
		
		// push pointers into vector
		groups.push_back( pg );		// add patch group to vector 	
		
		// no need to deallocate, since vector still pointing to the patch group
			
		}		// end of reading patches from file
	// all patch groups have been constructed
	printf("done loading patches\n\n");
	
	printf("using output texture size of %d\n\n", textSize);
	
	gamma = gam;
	
	
	printf( "finding co-planar patches that share edges\n\n" );
	
	shareEdges = new SimpleVector<int>[ numPatches ];
	
	for( int p=0; p<numPatches; p++ ) {
		// look at each other patch, check if it shares edges
		for( int q=0; q<numPatches; q++ ) {
			
			if( p != q ) {		// don't check patch against itself
			
				if( patches[p].areCoplanar( &(patches[q]) ) ) {		// check if coplanar
				
					// if so, check if vertex shared
					if( patches[p].shareVertex( &(patches[q]) ) ) {
						
						// make sure patches are same color before deciding that
						//  they share an edge (and are thus part of the same plane)
						
						if( red[p] == red[q] && green[p] == green[q] && blue[p] == blue[q] ) {		// check reflectance
							if( ered[p] == ered[q] && egreen[p] == egreen[q] && eblue[p] == eblue[q] ) {	// check emittence			
								// if yes, add to p's vector
								shareEdges[p].push_back( q );
								}	// end if emittence equal
							}	// end if reflectance equal
						}	// end check for shared vertex
					}	// end check for coplanar
				}	// end check for patches being identical
			}	// end for all p
		}	// end for all q
		
	// now know all patches that share edges (used when outputting textures)
	// can get rid of saved red, green, blue, and ered, egreen, eblue arrays
	
	delete [] red;
	delete [] green;
	delete [] blue;
	delete [] ered;
	delete [] egreen;
	delete [] eblue;
	
	
	// find largest patch area
	float largestWholePatchArea = 0;
	for( int p=0; p<numPatches; p++ ) {
		float area = patches[p].getArea();
		if( area > largestWholePatchArea ) {
			largestWholePatchArea = area;
			}
		}
	
	maxPatchArea = maxPArea;
	
	// do initial patch group subdivision
	// keep dividing until all patch groups contain patches of area less than threshold
	printf("Starting area-based patch subdivision, number of patch groups = %d\n", groups.size());
	
	// walk through group vector
	// keep checking size of vector as bounds, since vector will grow during this process
	int numSubdivisions = 0;
	int max = groups.size();
	for( int g=0; g<max; g++) {
		
		PatchGroup *thisPatchGroup = *(groups.getElement(g));
		
		if( thisPatchGroup->getPatchArea() > maxPatchArea ) {		// if patch too large
			numSubdivisions++;
			thisPatchGroup->forceSubdivide();
			char wasDivided = thisPatchGroup->subdivide( groups, g );
			
			delete thisPatchGroup;
			
			g--;	// stay at this spot in vector, since patch removed
			max = groups.size();
			}
		}
	printf("Done with initial patch subdivsion, number of patch groups = %d\n", groups.size());	
	
	
	// pass to output object
	outToFile = new Output( outFile, patches, numPatches, textSize, textSize, largestWholePatchArea, shareEdges );
	
	
	}
	
	
	
void Radiosity::run() {
	
	char restart = 1;		// have more subdivisions been made?
	
	int numRestarts = 0;
	
	while( restart ) {
			
		printf("Subdivision restart %d\n", numRestarts);
		numRestarts++;
		
		int numGroups = groups.size();
		
		printf("%d groups\n\n", numGroups);
		
		for( int g=0; g<numGroups; g++) {		// project all other group onto this group and calc form factors
			printf("Projecting patches onto group %d of %d\n", g, numGroups);
			(*(groups.getElement(g)))->project( groups );
			
			}
				
		
		// now start iterative radiosity calculation
		
		for( int c=0; c<3; c++) {		// for each color component
					
					/*
					*	Note:  Need to do in this order to save time and space.
					*	Reflective color component is multipled into components in form factor matrix.
					*	This should be done only once before the iteration starts
					*	However, there are three of these color components, and creating three 
					*	form factor matrices would take too much space
					*	So, must do colors one at a time.
					*
					*/
					
			float lastMaxDiff = iterThreshold + 1;		
		
			int numIter = 0;
			
			while( lastMaxDiff > iterThreshold ) {		// still more iterations needed
				printf( "Color %d, Iteration number %d \n", c, numIter );
				numIter++;
				
				lastMaxDiff = 0;
			
				for( int g=0; g<numGroups; g++) { // for each group
					PatchGroup *thisPatchGroup = *(groups.getElement(g));
					
					// get this patch's reflectance of this color component
					float reflectance = (*(thisPatchGroup->getReflectance()))[c];
					
					for(int p=0; p<4; p++) {	// for each patch in group

						//Color *patchRad = thisPatchGroup->getRadiosity(p);
						
						float *patchFormFactors = thisPatchGroup->getFormFactors(p);
						
						float componentSum = 0;
						
						// for each other patch in scene, add in contribution
						int otherPatchCount = 0;
						for( int gOther=0; gOther<numGroups; gOther++) { // for each other group
							PatchGroup *otherPatchGroup = *(groups.getElement(gOther));
							
							for(int pOther=0; pOther<4; pOther++) {	// for each patch in group
								
								Color *otherPatchRad = otherPatchGroup->getRadiosity(pOther);
								
								// only this color component
								float otherComponent = (*otherPatchRad)[c];
								
								// form factor of this patch times radiosity of other patch
								componentSum += patchFormFactors[otherPatchCount] * otherComponent;
								
								otherPatchCount++;
								}	// end for each patch in other group
							}	// end for each other group
						
						
						Color *patchRad = thisPatchGroup->getRadiosity(p);
						Color *patchEmittance = thisPatchGroup->getEmittance(p);
						
						float oldValue = (*patchRad)[c];
								
						(*patchRad)[c] = (*patchEmittance)[c] + reflectance * componentSum;	// set patch radiosity to new value 
				
						// check this change against lastMaxDiff
						float diff = (*patchRad)[c] - oldValue;
						if( diff > lastMaxDiff ) lastMaxDiff = diff;
						else if( -diff > lastMaxDiff) lastMaxDiff = -diff;
				
						}	// end for each patch in group
			
					}	// end for each group
		
		
		
				}
					
				
			}		// end each color
		
		// subdivide, and test if subdivisions were made
	
		restart = false;
		for( int g=0; g<numGroups; g++) { // for each group
			PatchGroup *thisPatchGroup = *(groups.getElement(g));
			
			// check patch and mark it for subdivision
			thisPatchGroup->checkThreshold( subDivideThreshold );
			}
		
		// now carry out subdivisions and removal
		//  subtle, because groups are being removed from vector as this happens
		for( int g=0; g<numGroups; g++) { // for each group
			PatchGroup *thisPatchGroup = *(groups.getElement(g));
			
			char wasDivided = 0;
			
			// let this patch sub divide and remove itself
			wasDivided = thisPatchGroup->subdivide( groups, g);
			
			if( wasDivided ) {		// patch divided and removed
				// delete this group, no longer needed, not in vector
				delete thisPatchGroup;
				g = g-1;	// stay at this spot in vector
				numGroups = numGroups -1;	// fewer old groups in vector
				restart = 1;	// flag that some have been divided, need to restart
				}
			}
			
		if( restart ) {
			// need to clear all radiosities before restart
			int numGroups = groups.size();
			for( int g=0; g<numGroups; g++) { // for each group
				PatchGroup *thisPatchGroup = *(groups.getElement(g));
			
				// clear this patch's radiosity
				thisPatchGroup->clearRadiosity();
				}
			
			}
			
		}		// end while restarting


	// at this point, radiosity has satisfied both thresholds
	//	differences across patch subdivisions are less than *subDivideThreshold*
	//	differences between iterations are less than *iterThreshold*

	int numGroups = groups.size();
	
	// normallize colors in scene
	
	// first, gamma correct patches
	if( gamma != 1.0 ) {
		for( int g=0; g<numGroups; g++) { // for each group
			PatchGroup *thisPatchGroup = *(groups.getElement(g));
			for( int p=0; p<4; p++ ) {	// for each sub patch in group
				for( int c=0; c<3; c++ ) {	// for each color
					float color = (*(thisPatchGroup->getRadiosity(p)))[c];
					color = pow(color, gamma);
					(*(thisPatchGroup->getRadiosity(p)))[c] = color;
					}	
				
				}
			}
		}
	
	// second, walk through all sub patches, find maximum color compnent
	float max = -FLT_MAX;
	for( int g=0; g<numGroups; g++) { // for each group
		PatchGroup *thisPatchGroup = *(groups.getElement(g));
		for( int p=0; p<4; p++ ) {	// for each sub patch in group
			float subPatchMaxComponent = thisPatchGroup->getRadiosity(p)->getMax();
			if( subPatchMaxComponent > max ) {
				max = subPatchMaxComponent;
				}
			}
		}
	// third, weight all sub patches by max inverse to normaliz
	float invMax = 1/max;
	for( int g=0; g<numGroups; g++) { // for each group
		PatchGroup *thisPatchGroup = *(groups.getElement(g));
		for( int p=0; p<4; p++ ) {	// for each sub patch in group
			thisPatchGroup->getRadiosity(p)->weightColor( invMax );
			}
		}		
	// now all colors normalized, and color composites up to date
	
	
	// now output to disk
	printf( "\n\nOutputing to disk...\n" );
	outToFile->buildFromGroups( groups );
	printf( "Done\n" );
	}

	

	