// Jason Rohrer
// Runner.cpp

/**
*
*	File for main method of RadiosGL
*	Gathers command line args, runs radiosity
*
*	Created 10-24-99
*	Mods:
*		Jason Rohrer	12-1-99		Added net time monitor.
*		Jason Rohrer	2-22-2000	Added support for reading params from a file
*									(if no command line args specified)
*
*/


#include <stdlib.h>
#include <time.h>

#include "Radiosity.h"


// prototypes
void usage();
void unknownOption(char *s);
void unknownArgument(char *s);
void testRun();


char *paramsFileName = "RadiosGL.ini";


int main( int argc, char *argv[] ) {
	
	printf(".oO()   HardenedCriminal Software (TM)   ()Oo.\n\n");
	printf("RadiosGL Package  (c)2000 Little  (open source, end to tyranny)\n");
	printf("RadiosGL:  command line radiosity calculator\n");
	printf("Version 1.0\n\n");
	
	// out file is optional
	FILE *outFile;
	FILE *modlFile;
	FILE *hcFile;
	
	char *outFileName = "RadiosGL.out";
	int outFileArgNumber = 0;
	int modlFileArgNumber = 0;
	int hcFileArgNumber = 0;
	
	float hemicubeWorldWide = 0.0;
	float maxPatchArea = 0.0;
	float iterThreshold = 0.0;
	float subDivideThreshold = 0.0;
	float gamma = 1.0;
	int textSize = 64;
	
	int clearArgV = false;		// argv needs deallocation?
	
	if( argc == 0 ) {	// read params from a file
		printf( "No input arguments specified, trying read them from file:  %s\n\n", paramsFileName );
		// try opening the file
		FILE *paramsFile = fopen( paramsFileName, "r" );	// a text file
		
		if( paramsFile == NULL ) {	// file open failed
			printf("Could not open file for reading:  %s\n\n", paramsFileName);
			usage();
			}
		
		
		// read params from file, stick them into arg array
		
		delete [] argv;
		
		argv = new char *[99];		// make it big enough to hold args
		
		int retVal = 1;
		int v = 0;
		
		// note:  skip first argument slot, since this is where name of executable is kept
		
		while( retVal != 0 && retVal != EOF ) {
			v++;
			argv[v] = new char[99];
			retVal = fscanf( paramsFile, "%s", argv[v] );
			}
		// last v-index in argv is not valid arg, so v = number of args in argv
		argc = v;
		
		clearArgV = true;		// set flag to deallocate argv memory
		}
		
		
		
	
	if( argc < 12 ) {		// must at least specify 2 files and 4 numbers (with 2 params each (one for switch))
		usage();
		}
	
	// argc >= 12
	int i=1;
	
	while( i<argc ) {
		switch( argv[i][0] ) {
			case '-':			// read any switches
				switch( argv[i][1] ) {
					case '?':
						usage();
						break;
					case 'm':
						i++;
						modlFileArgNumber = i;
						i++;				
						break;
					case 'h':
						i++;
						hcFileArgNumber = i;
						i++;				
						break;
					case 'o':
						i++;
						outFileArgNumber = i;
						i++;
						break;
					case 'w':
						i++;
						sscanf(argv[i], "%f", &hemicubeWorldWide);
						i++;
						break;
					case 'a':
						i++;
						sscanf(argv[i], "%f", &maxPatchArea);
						i++;
						break;
					case 'i':
						i++;
						sscanf(argv[i], "%f", &iterThreshold);
						i++;
						break;
					case 's':
						i++;
						sscanf(argv[i], "%f", &subDivideThreshold);
						i++;
						break;
					case 'g':
						i++;
						sscanf(argv[i], "%f", &gamma);
						i++;
						break;
					case 't':
						i++;
						sscanf(argv[i], "%d", &textSize);
						i++;
						break;
					default:		// handle bad switches
						unknownOption(argv[i]);
						break;
					}
				break;
			default:			// handle bad arguments
				unknownArgument(argv[i]);
				i++;
				break;
			}
		}

	if( hemicubeWorldWide <= 0 ) {
		printf("hemicube_world_wide must be greater than 0\n\n");
		usage();
		}
	if( maxPatchArea <= 0 ) {
		printf("max_patch_area must be greater than 0\n\n");
		usage();
		}
	if( iterThreshold <= 0 ) {
		printf("iteration_threshold must be greater than 0\n\n");
		usage();
		}
	if( subDivideThreshold <= 0 ) {
		printf("subdivide_threshold must be greater than 0\n\n");
		usage();
		}
	if( textSize <= 0 ) {
		printf("texture_size must be greater than 0\n\n");
		usage();
		}

	// params read correctly
	
	// open out file for binary writing
	if( outFileArgNumber == 0 ) {			// default name
		outFile = fopen((const char*)outFileName, "wb");
		if( outFile == NULL ) {
			printf("Could not open file for writing:  %s\n\n", outFileName);
			usage();
			}
		}
	else {					// specified name
		outFile = fopen((const char*)argv[outFileArgNumber], "wb");
		if( outFile == NULL ) {
			printf("Could not open file for writing:  %s\n\n", argv[outFileArgNumber]);
			usage();
			}
		}

	if( modlFileArgNumber == 0) {
		printf("Must specify model_file\n\n");
		usage();
		}
	else {				// open specified model file in text mode for reading
		modlFile = fopen((const char*)argv[modlFileArgNumber], "r");
		if( modlFile == NULL ) {
			printf("Could not open file for reading:  %s\n\n", argv[modlFileArgNumber]);
			usage();
			}
		}
	
	if( hcFileArgNumber == 0) {
		printf("Must specify hemicube_file\n\n");
		usage();
		}
	else {				// open specified hemicube file in binary mode for reading
		hcFile = fopen((const char*)argv[hcFileArgNumber], "rb");
		if( hcFile == NULL ) {
			printf("Could not open file for reading:  %s\n\n", argv[hcFileArgNumber]);
			usage();
			}
		}


	// all files open, all number inputs parsed:

	// clear argv if need be
	if( clearArgV ) {
		for( int p=0; p<argc; p++ ) {
			delete [] argv[p];
			}
		delete [] argv;
		}

	
	// start timing
	unsigned long startTime = (unsigned)time( NULL );
	unsigned long netTime;
	
	// create radiosity object
	Radiosity mainRadiosity(modlFile, hcFile, outFile, hemicubeWorldWide, maxPatchArea, iterThreshold, subDivideThreshold, gamma, textSize);
		
	// run radiosity
	mainRadiosity.run();
	
	
	// end timing
	netTime = (unsigned)time( NULL ) - startTime;
	printf("\n%d seconds elapsed.\n", netTime);
	
	
	
	return 0;
	}
	
	
void usage() {				// print usage help message and exit

	printf("usage:\n");
	printf("RadiosGL\n");
	printf("\t[-?]\n\t");
	printf("-m model_file  -h hemicube_file\n\t");
	printf("[-o output_file]  -w hemicube_world_wide\n\t");
	printf("-a max_patch_area\n\t");
	printf("-i iteration_threshold\n\t");
	printf("-s subdivide_threshold\n\t");
	printf("[-g gamma]  [-t texture_size]\n\n");
	
	printf("examples:\n\t");
	printf("RadiosGL -m test.mdl -h myHemicube.hc -w 2.3 -a 50.0 -i 0.01 -s 0.05 -t 64 \n\n\t");
	printf("RadiosGL -m test.mdl -h myHemicube.hc -w 2.3 -a 50.0 -i 0.01 -s 0.05 -g 0.25 \n\n\t");
	printf("RadiosGL -m test.mdl -o test.out -h myHemicube.hc -w 2.3 -a 50.0 -i 0.01 -s 0.05 \n\n\t");
	printf("RadiosGL\n\n\t");
	
	printf("(if no parameters specified, parameters are read from %s file)\n\n", paramsFileName);
	
	exit(0);
	}
	
void unknownOption(char *s) {
	printf("unknown option:  %s\n\n", s);
	usage();
	}
	
void unknownArgument(char *s) {
	printf("unknown argument:  %s\n\n", s);
	usage();
	}