// Jason Rohrer
// Vertex.h

/**
*
*	Vertex object for RadiosGL
*
*	Float coords in X, Y, Z
*
*
*	Created 9-5-99
*	Mods:
*
*		Jason Rohrer	12-1-99		Fixed horrid memory leaks:
*									For functions that return a Vertex (not a pointer)
*									it used to creat "new Vertex()" and then return
*									a dereferenced version of this.  This was being
*									allocated on the heap, and these vertex objects
*									never got dealocated!
*									Optimization:
*									Added overloaded version of rotations that
*									use precalculated sine and cosine hints
*
*		Jason Rohrer	1-16-2000	Added length function for when vertex is treated
*									as a vector
*
*
*/

#ifndef VERTEX_INCLUDED
#define VERTEX_INCLUDED

#include <math.h>
#include <stdio.h>
#include "Color.h"
#include "RadiosityConstants.h"


class Vertex {
	public :
		Vertex(float inX, float inY, float inZ);
		Vertex();
		float x, y, z;
		
		void move(Vertex &deltaVert);	// move by delta
		
		void rotateX(float thetaX);		// rotate around x axis
		void rotateY(float thetaY);		// rotate around y axis
		void rotateZ(float thetaZ);		// rotate around z axis
		
		void rotateX( float sinTheta, float cosTheta );		// rotate around x axis, w/ hint
		void rotateY( float sinTheta, float cosTheta );		// rotate around y axis, w/ hint
		void rotateZ( float sinTheta, float cosTheta );		// rotate around z axis, w/ hint
		
		Vertex copy();			// make a copy of this vertex
		
		// calculate angles of rotation needed to line vertex up with Z axis
		// note:  rotation X, THEN rotation Y order must be executed to line up with Z
		void calcRotationToZ( float &thetaX, float &thetaY );
		
		void print();	// print coords for this vertex
		
		float length();		// return distance of vertex from origin
							// (useful as vector length)	
	};
	
	

inline Vertex::Vertex(float inX, float inY, float inZ) {
	x = inX;
	y = inY;
	z = inZ;	
	}

inline Vertex::Vertex() {
	x=0;
	y=0;
	z=0;
	}


inline void Vertex::move(Vertex &deltaVert) {
	x += deltaVert.x;
	y += deltaVert.y;
	z += deltaVert.z;
	}

// inline rotation functions

inline void Vertex::rotateX( float sinTheta, float cosTheta ) {
	float oldY = y;
	y = y * cosTheta - z * sinTheta;
	z = oldY * sinTheta + z * cosTheta;
	}
	
inline void Vertex::rotateY( float sinTheta, float cosTheta ) {
	float oldX = x;
	
	x = cosTheta * x + sinTheta * z;
	z = -sinTheta * oldX + cosTheta * z;
	}
		
inline void Vertex::rotateZ( float sinTheta, float cosTheta ) {
	float oldX = x;
	
	x = cosTheta * x - sinTheta * y;
	y = sinTheta * oldX + cosTheta * y;
	}

inline void Vertex::rotateX(float thetaX) {
	float cosTheta = cos(thetaX);
	float sinTheta = sin(thetaX);
	rotateX( sinTheta, cosTheta );
	}
	
inline void Vertex::rotateY(float thetaY) {
	float cosTheta = cos(thetaY);
	float sinTheta = sin(thetaY);
	rotateY( sinTheta, cosTheta );
	}
		
inline void Vertex::rotateZ(float thetaZ) {
	float cosTheta = cos(thetaZ);
	float sinTheta = sin(thetaZ);
	rotateZ( sinTheta, cosTheta );
	}

inline float Vertex::length() {
	return sqrt(x * x + y * y + z * z);
	}

inline Vertex Vertex::copy() {
	Vertex returnVert(x, y, z);
	return returnVert;
	}


inline void Vertex::calcRotationToZ( float &thetaX, float &thetaY ){
	Vertex tempNorm( x, y, z );	
	
	// FIRST:  calculate rotation around X axis
	float tanThetaX = 0;
	
	thetaX = 0;
	
	if( tempNorm.z == 0 ) {
		if( tempNorm.y > 0 ) {
			thetaX = PI * .5;
			}
		else {
			thetaX = - PI * .5;
			}
		}
	else {
		tanThetaX = tempNorm.y / -tempNorm.z;
		
		thetaX = -atan( tanThetaX );
		
		if( tempNorm.z < 0 ) {
			thetaX += PI;
			}
		}
	
	//rotate a temp normal by this before computing thetaY
	tempNorm.rotateX( thetaX );
	
	
	// SECOND:  calculate rotation around Y axis needed after X rotation
	float tanThetaY = 0;
	
	thetaY = 0;
	
	if( tempNorm.z == 0 ) {
		if( tempNorm.x > 0 ) {
			thetaY = - PI * .5;
			}
		else {
			thetaY = PI * .5;
			}
		}
	else {
		tanThetaY = tempNorm.x / tempNorm.z;

		thetaY = -atan( tanThetaY );
		
		if( tempNorm.z < 0 ) {
			thetaY += PI;
			}
		}
	}


inline void Vertex::print() {
	printf("(%f, %f, %f)", x, y, z);
	}


// implement inline operators in namespace
	
inline Vertex operator + (Vertex v, Vertex w) {
	Vertex returnVert(v.x + w.x, v.y + w.y, v.z + w.z);
	return returnVert;
	}

inline Vertex operator - (Vertex v, Vertex w) {
	Vertex returnVert(v.x - w.x, v.y - w.y, v.z - w.z);
	return returnVert;
	}

inline Vertex operator * (Vertex v, float f) {
	Vertex returnVert(v.x * f, v.y * f, v.z * f);
	return returnVert;
	}
	
inline Vertex operator * (float f, Vertex v) {
	Vertex returnVert(v.x * f, v.y * f, v.z * f);
	return returnVert;
	}


inline Vertex operator / (Vertex v, float f) {
	float invF = 1/f;
	Vertex returnVert(v.x * invF, v.y * invF, v.z * invF);
	return returnVert;
	}


inline float dot(Vertex a, Vertex b) {
	return a.x * b.x + a.y * b.y + a.z * b.z;
	}





	
#endif