// Jason Rohrer
// ViewCube.cpp

/**
*
*	ViewCube object for RadiosGL
*
*	patch specific hemicube "instance"  (points to global hemicube)
*
*
*	Created 9-17-99
*	Mods:
*			Jason Rohrer	9-26-99		moved z-buffering to zBufferPatch.h
*			Jason Rohrer	10-31-99	fixed patch projection formulas
*			Jason Rohrer	12-1-99		Optimization:
*										Precalculate many values
*/



#include "ViewCube.h"




// constructor
ViewCube::ViewCube(Vertex cVert, Vertex N, Hemicube *h) {
	centVert = cVert;
	norm = N;
	hCube = h;
	
	numPixelsWide = hCube->getPixelsWide();
	
	
	// init arrays
	int numPixelsTop = numPixelsWide * numPixelsWide;
	
	topPixels = new ZPixel[ numPixelsTop ];
	
	int numPixelsSide = numPixelsTop / 2;
	
	for( int i=0; i<4; i++) {
		sidePixels[i] = new ZPixel[ numPixelsSide ];
		}
	
	worldToPixUnit = numPixelsWide / hCube->getWorldWide();
		
	// precalc y offsets into pixel planes
	yOffset = new int[numPixelsWide];
	
	for( int y=0; y<numPixelsWide; y++ ) {
		yOffset[y] = numPixelsWide * y;
		}
		
	// precalculate several values
		Vertex cubeCenter(0,0,0);
		deltaCenter = cubeCenter - centVert;
		
		// calculate rotation needed to line normal up with Z+
		thetaX = 0;
		thetaY = 0;
		norm.calcRotationToZ( thetaX, thetaY );
		
		halfWorldWide = (hCube->getWorldWide()) * 0.5;
		startXWorld = - halfWorldWide;
		endXWorld = halfWorldWide;
		startYWorld = - halfWorldWide;
		endYWorld =  halfWorldWide;
		startSideYWorld = 0;
	}
	
ViewCube::~ViewCube() {
	delete [] topPixels;
	for( int i=0; i<4; i++) {
		delete [] sidePixels[i];
		}
	delete [] yOffset;
	}
	
	

void ViewCube::projPatchOnCube(int pID, Patch *pIn) {
	
	// make a copy of the patch to preserver original
	Patch *p = pIn->copy();
	
	// for top:		(special case)
			

		// move patch so that cube center is at (0,0,0)
		p->move(deltaCenter);		// move the patch


		// Now rotate patch so that cube normal lines up with Z+
		p->rotateX(thetaX);	
		p->rotateY(thetaY);
		
					
		// add this patch to the top z-buffer
		zBufferPatch(topPixels, yOffset, p, pID, numPixelsWide, numPixelsWide,
							startXWorld, endXWorld, 
							startYWorld, endYWorld,
							halfWorldWide);
								
			
	// now rotate patch and cube so that side[0] normal points along Z+
	// rotate clockwise 90 degrees around y axis
	
	float delTheta = halfPI;
	p->rotateY(-delTheta);
	p->rotateZ(-delTheta);		// now x and y dirs on side[0] match world x, y dirs
	
	int yPixWide = numPixelsWide>>1;
	
	// note:  can use the same y offsets for top and sides
	
	// add this patch to side[0] z-buffer
	zBufferPatch(sidePixels[0], yOffset, p, pID, numPixelsWide, yPixWide,
						startXWorld, endXWorld,
						startSideYWorld, endYWorld,
						halfWorldWide);		
				
	for( int s=1; s<4; s++) {		// for each remaining side
		
		// rotate 90 deg. in cw y direction to make side face z+
		
		p->rotateY(-delTheta);
		// add this patch to side[s] z-buffer
		zBufferPatch(sidePixels[s], yOffset, p, pID, numPixelsWide, yPixWide,
						startXWorld, endXWorld,
						startSideYWorld, endYWorld,
						halfWorldWide);
		}
	
	// delete the copy of the patch
	delete p;
		
	}	// end projPatchOnCube()
	
	
	
	
void ViewCube::sumFormFactors(float *formFactors) {
		
	// look in top of cube
	int pixWideTop = hCube->getPixelsWide();
	int pixInTop = pixWideTop * pixWideTop;
		
	for( int pix=0; pix<pixInTop; pix++) {
		int id = topPixels[pix].getObjectID();
		// if an object is in buffer at this point, and it is facing the buffer
		if( id >= 0 && !(topPixels[pix].getIsBack()) ) {		
			formFactors[id] += hCube->getFormFactor(0, pix, -1);
			}
		}
			
	int pixHighSide = pixWideTop >> 2;		
	int pixInSide = pixHighSide * pixWideTop;
	
	for( int s=0; s<4; s++) {	// look in each cube side
		for( int pix=0; pix<pixInSide; pix++) {
			int id = sidePixels[s][pix].getObjectID();
			// if an object is in buffer at this point, and it is facing the buffer
			if( id >= 0 && !(sidePixels[s][pix].getIsBack()) ){
				formFactors[id] += hCube->getFormFactor(s+1, pix, -1);
				}
			}
		}	
	
	}	// end sumFormFactors()
	
	
void ViewCube::print() {
	printf("view cube top z buffer contents:\n");
	for( int y=0; y<numPixelsWide; y++) {
		int yOffset = y * numPixelsWide;
		for(int x=0; x<numPixelsWide; x++) {
			int val = topPixels[yOffset + x].getObjectID();
			if(val == -1) printf(" x  "); 
			else { 
				printf("%2d", val);
				if( topPixels[yOffset + x].getIsBack() ) {
					printf("b ");
					}
				else{
					printf("  ");
					}	
				}
			}
		printf("\n");
		}
		
	for( int s=0; s<4; s++ ) {		// side number to print	
		printf("view cube side %d z buffer contents:\n", s);
		int pixHighSide = numPixelsWide >> 2;
		for( int y=0; y<pixHighSide; y++) {
			int yOffset = y * numPixelsWide;
			for(int x=0; x<numPixelsWide; x++) {
				int val = sidePixels[s][yOffset + x].getObjectID();
				if(val == -1) printf(" x  "); 
				else { 
					printf("%2d", val);
					if( topPixels[yOffset + x].getIsBack() ) {
						printf("b ");
						}
					else{
						printf("  ");
						}	
					}	
				}
			printf("\n");
			}
		}
		
	printf("\n\n");
	}