// Jason Rohrer
// MakeHemicube.cpp

/**
*
*	Outputs a hemicube file for RadiosGL
*
*
*	Created 9-12-99
*	Mods:
*		Jason Rohrer	9-17-99		Fixed form factor formulas.  Fixed command line
*									argument handling
*
*		Jason Rohrer	1-16-2000	Changed to use UniversalFileIO to work on both big 
*									and little endian machines									
*
*/



#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "UniversalFileIO.h"


// prototypes
void usage();
void unknownOption(char *s);
void unknownArgument(char *s);
void writeCubeData( FILE *outFile, int cubeWidth);


char *paramsFileName = "MakeHemicube.ini";


int main( int argc, char *argv[] ) {

	UniversalFileIO fileIO;

	int cubeWidth = 0;
	
	FILE *outFile;
	
	char *outFileName = "hemicube.out";
	int outFileArgNumber = 0;
	
	int clearArgV = false;
	
	
	if( argc == 0 ) {	// read params from a file
		printf( "No input arguments specified, trying read them from file:  %s\n\n", paramsFileName );
		// try opening the file
		FILE *paramsFile = fopen( paramsFileName, "r" );	// a text file
		
		if( paramsFile == NULL ) {	// file open failed
			printf("Could not open file for reading:  %s\n\n", paramsFileName);
			usage();
			}
		
		
		// read params from file, stick them into arg array
		
		delete [] argv;
		
		argv = new char *[99];		// make it big enough to hold args
		
		int retVal = 1;
		int v = 0;
		
		// note:  skip first argument slot, since this is where name of executable is kept
		
		while( retVal != 0 && retVal != EOF ) {
			v++;
			argv[v] = new char[99];
			retVal = fscanf( paramsFile, "%s", argv[v] );
			}
		// last v-index in argv is not valid arg, so v = number of args in argv
		argc = v;
		
		clearArgV = true;		// set flag to deallocate argv memory
		}
	
	
	if( argc <2 ) {		// must at least specify cube size
		usage();
		}
	
	// argc >= 2
	
	int i=1;
	
	while( i<argc ) {
		switch( argv[i][0] ) {
			case '-':			// read any switches
				switch( argv[i][1] ) {
					case '?':
						usage();
						break;
					case 'o':
						i++;
						outFileArgNumber = i;
						i++;
						break;
					default:
						unknownOption(argv[i]);
						break;
					}
				break;
			default:			// read size of cube number
				int numRead = sscanf(argv[i], "%d", &cubeWidth);
				if( numRead == 0 ) {
					unknownArgument(argv[i]);
					}
				i++;
				break;
			}
		}
	
	if( cubeWidth <= 0 ) {
		printf("pixel_width_of_cube must be greater than 0\n\n");
		usage();
		}
	
	// params read correctly
	
	// clear argv if need be
	if( clearArgV ) {
		for( int p=0; p<argc; p++ ) {
			delete [] argv[p];
			}
		delete [] argv;
		}

	
	
	
	if( outFileArgNumber == 0 ) {
		outFile = fopen((const char*)outFileName, "wb");
		if( outFile == NULL ) {
			printf("Could not open file for writing:  %s\n\n", outFileName);
			usage();
			}
		}
	else {
		outFile = fopen((const char*)argv[outFileArgNumber], "wb");
		if( outFile == NULL ) {
			printf("Could not open file for writing:  %s\n\n", argv[outFileArgNumber]);
			usage();
			}
		}					
	
	
	
	// params okay, write the file
	
	// header
	fileIO.fwriteLong( outFile, cubeWidth );
	
	writeCubeData( outFile, cubeWidth );	// data body	
	
	fclose( outFile );
	
	
	printf("Done.\n");
	
	return 0;
	}
	

void usage() {				// print usage help message and exit

	printf("usage:\n");
	printf("MakeHemicube\n");
	printf("\t[-?]\n\t");
	printf("[-o output_file] pixel_width_of_cube\n\n");
	
	printf("example:\n\t");
	printf("MakeHemicube -o myHemicube 1000\n\n");
	
	exit(0);
	}
	
void unknownOption(char *s) {
	printf("unknow option:  %s\n\n", s);
	usage();
	}
	
void unknownArgument(char *s) {
	printf("unknow argument:  %s\n\n", s);
	usage();
	}



void writeCubeData( FILE *outFile, int cubeWidth) {
	UniversalFileIO fileIO;
	
	float PI = 3.14159265358979323846;
	
	
	int sizeOfRegion = cubeWidth / 2;	// = size of upper left corner of cube top
										// = size of left half of cube side 
	
	int pixInRegion = sizeOfRegion * sizeOfRegion;
	
	float *regionBuffer = new float[pixInRegion];	 // region buffer, filled and flushed to file twice
	
	// first region, upper left corner of cube top
	for( int buffY= sizeOfRegion - 1; buffY>=0; buffY--) {
		for( int buffX = 0; buffX < sizeOfRegion; buffX++) {
			float x = buffX - sizeOfRegion - 0.5;	// fix center of pixel error with 0.5 offset
			float y = buffY + 0.5;
			float temp = x*x + y*y + sizeOfRegion * sizeOfRegion;
			temp = temp * temp;
				
			regionBuffer[buffY * sizeOfRegion + buffX] = (sizeOfRegion * sizeOfRegion) / (PI * temp);
			}
		}
	
	for( int p=0; p<pixInRegion; p++ ) {
		fileIO.fwriteFloat( outFile, regionBuffer[p] );
		}
	
	// second region, left half of cube side
	for( int buffY= sizeOfRegion - 1; buffY>=0; buffY--) {
		for( int buffX = 0; buffX < sizeOfRegion; buffX++) {
			float x = buffX - sizeOfRegion - 0.5;	// fix center of pixel error with 0.5 offset
			float y = buffY + 0.5;
			float temp = x*x + sizeOfRegion * sizeOfRegion;
			float temp2 = temp + y*y;
			temp2 = temp2 * temp2;
			temp = sqrt(temp);
				
			regionBuffer[buffY * sizeOfRegion + buffX] = (y * temp) / (PI * temp2);
			}
		}
	
	for( int p=0; p<pixInRegion; p++ ) {
		fileIO.fwriteFloat( outFile, regionBuffer[p] );
		}
	
	
	
	return;
	}

	