// Jason Rohrer
// zBufferPatch.cpp

/**
*
*	function for z-buffering a patch onto a viewplane
*
*
*
*	Created 9-24-99
*	Mods:
*		Jason Rohrer	10-31-99		Added support for back of patch detection
*
*		Jason Rohrer	11-21-99		Changed projection model slightly
*										Patches that are partially behind the projection point
*											(e.g., crossing the xy plan and changing sign of z)
*											are now clipped entirely
*										They will likely be caught and projected later once they
*											are subdivided into patches, some of which will
*											be in front of the projection point
*
*		Jason Rohrer	11-29-99		Clipped patches are not always being caught in later subdivisions
*											since the patches being projected may not get subdivided.
*											Fixed to actually clip vertices with negative z instead of dropping patch
*											entirely.  Approximates clipping without creating a quad polygon
*											(moves bad triangle vertex inbounds)
*		Jason Rohrer	12-1-99			Optimization:
*										Precalculate array index in innermost loop
*
*		Jason Rohrer	12-3-99			Fixed problem in generalization that showed up with hemicube sides:
*										Point of projection should always be at (0,0,0), not centered
*										with respect to the image plane.  (for sides of the hemicube, the
*										point of projection was ending up too far towards y+)
*
*/

#include "zBufferPatch.h"


void zBufferPatch( ZPixel *viewPlane, int *yOffset, Patch *p, int pID, 
						int planePixWide, int planePixHigh,
						float planeStartX, float planeEndX, 
						float planeStartY, float planeEndY,
						float d) {
	
	float minDist = 0.001 * d;	// near clipping plane
	
	char isBack = false;
	
	// test whether patch back is facing this viewplane
	
	Vertex patchNorm = p->getNormal();
	Vertex planeNorm(0,0,1);
	if( dot(patchNorm, planeNorm) >0 ) {		// patch facing away from view plane
		isBack = true;
		}
	
	// now project patch onto plane
	// ys = (y * d)/z
	// xs = (x * d)/z
	// zs = -1/z
			
	char areZValsBehind = 0;		// does polygon have all three verts with negative z?				
			
	Patch *projected = p->copy();
			
	for(int i=0; i<3; i++) {	// calculate projected coords
		
		
		if( projected->vertices[i].z < minDist ) {		// this vertex needs to be clipped
			// walk along edge towards +z to find a new vertex
			
			int nextInd = 1;
			
			// walk towards other vertex with greatest Z value
			// 	( this lets patches that share vertices remain in contact )
			float maxZ = -1;
			int indOfMaxZ = 0;
			
			for( nextInd = 1; nextInd < 3; nextInd++ ) {
				float thisZ = projected->vertices[(i+nextInd)%3].z;
				if( thisZ > maxZ ) {
					maxZ = thisZ; 
					indOfMaxZ = nextInd;
					}
				}
				
			nextInd = indOfMaxZ;	
			if( maxZ < minDist ) {		// check if patch is entirely clipped
				areZValsBehind = 1;
				}
						
			// nextInd points to a vertext with non-negative z value
			
			if( !areZValsBehind ) {
				float deltaX = projected->vertices[(i+nextInd)%3].x - projected->vertices[i].x;
				float deltaY = projected->vertices[(i+nextInd)%3].y - projected->vertices[i].y;
				float deltaZ = projected->vertices[(i+nextInd)%3].z - projected->vertices[i].z;
				
				float invDeltaZ = 1 / deltaZ;
				float slopeX = deltaX * invDeltaZ;
				float slopeY = deltaY * invDeltaZ;
				
				float deltaToZero = minDist - projected->vertices[i].z;
				
				// new x and y of vert created by clipping
				projected->vertices[i].x = deltaToZero * slopeX + projected->vertices[i].x;
				projected->vertices[i].y = deltaToZero * slopeY + projected->vertices[i].y;
			
				// new z is 0
				projected->vertices[i].z = minDist;
				}
			}
		
		float invZ = 1 / projected->vertices[i].z;
		
		projected->vertices[i].x = (projected->vertices[i].x * d) * invZ;
		projected->vertices[i].y = (projected->vertices[i].y * d) * invZ;
		projected->vertices[i].z = -invZ;
		}		
	
	Vertex *vert = projected->vertices;
	
		
	// make sure this polygon has some verts with positive z vals
	if( ! areZValsBehind ) {
				
		int *maxRowX = new int[planePixHigh];
		int *minRowX = new int[planePixHigh];
		float *zOfMaxRowX = new float[planePixHigh];
		float *zOfMinRowX = new float[planePixHigh];
					
					
		for( int i=0; i<planePixHigh; i++) {
			maxRowX[i] = -planePixWide;
			minRowX[i] = planePixWide;
			zOfMaxRowX[i] = 1;
			zOfMinRowX[i] = 1;
			}
					
		
		float worldToPixUnit = planePixWide/(planeEndX - planeStartX);
		
		int planeStartXInPix = (int)(worldToPixUnit * planeStartX);		// start and end of view plane in pixels
		int planeEndXInPix = planeStartXInPix + planePixWide -1;
		
		int planeStartYInPix = (int)(worldToPixUnit * planeStartY);
		int planeEndYInPix = planeStartYInPix + planePixHigh -1;
		
		int yPixArrayOffset = 0 - planeStartYInPix;		// how much offset needed to get into positive array indices
		int xPixArrayOffset = 0 - planeStartXInPix;
		
					
		for( int i=0; i<3; i++) {		// for each side of polygon
			int startXInPix = (int)(vert[i].x * worldToPixUnit);
			int startYInPix = (int)(vert[i].y * worldToPixUnit);
			int endXInPix = (int)(vert[(i+1)%3].x * worldToPixUnit);
			int endYInPix = (int)(vert[(i+1)%3].y * worldToPixUnit);
			
			float startZ = vert[i].z;
			float endZ = vert[(i+1)%3].z;
			
			
			
			// make sure y increasing along edge
			if( startYInPix > endYInPix) {
				int temp = startYInPix;
				startYInPix = endYInPix;
				endYInPix = temp;
				
				temp = startXInPix;
				startXInPix = endXInPix;
				endXInPix = temp;
				
				float tempf = startZ;
				startZ = endZ;
				endZ = tempf;
				}

			// watch for case where side is outside y range of viewplane
			if( endYInPix > planeStartYInPix && startYInPix < planeEndYInPix ) {

				// find max and min x contributions from this edge	
			
				float deltaX = endXInPix - startXInPix;
				float deltaY = endYInPix - startYInPix;
				float deltaZ = endZ - startZ;
						
				float invDeltaY = 1/deltaY;
						
				float slopeX = deltaX * invDeltaY;
				float slopeZ = deltaZ * invDeltaY;
						
				int realStartY = startYInPix;
				int realEndY = endYInPix;
						
				if( startYInPix < planeStartYInPix )  realStartY = planeStartYInPix;	
				if( endYInPix > planeEndXInPix )  realEndY = planeEndYInPix;
						
				int currentY = realStartY;
				float currentX = slopeX * (realStartY - startYInPix) + startXInPix;
				float currentZ = slopeZ * (realStartY - startYInPix) + startZ;
					
				// while in bounds
				while( currentY <= realEndY ) {
					
					int arrayIndex = currentY + yPixArrayOffset;
						
					// compare x against max for this row
					if( maxRowX[ arrayIndex ] < currentX ) {
						maxRowX[ arrayIndex ] = (int)(currentX);
						zOfMaxRowX[ arrayIndex ] = currentZ;
						}
					// compare x against min for this row
					if( minRowX[ arrayIndex ] > currentX ) {
						minRowX[ arrayIndex ] = (int)(currentX);
						zOfMinRowX[ arrayIndex ] = currentZ;
						}
					currentY++;
					currentZ += slopeZ;
					currentX += slopeX;
					}	// end while
				}	// end if in y range
			}	// end for all sides of poly
					
					
					
		// now have max and min X for each row, with Z values.... fill in the Z buffer
		
		
		// clip max and min values for each row 
		for( int y=0; y<planePixHigh; y++ ) {
			float startZ = zOfMinRowX[y];
			float endZ = zOfMaxRowX[y];
			float deltaZ = endZ - startZ;
			int deltaPix = maxRowX[y] - minRowX[y];
			float slopeZ = deltaZ/deltaPix;
			
			if( maxRowX[y] > planeEndXInPix ) {		// need to clip this max
				
				float deltaX = (float)(maxRowX[y] - planeEndXInPix);
				zOfMaxRowX[y] = zOfMaxRowX[y] - deltaX * slopeZ;
				
				maxRowX[y] = planeEndXInPix;
				}
			 
			 if( minRowX[y] < planeStartXInPix ) {		// need to clip this min
				
				float deltaX = (float)(minRowX[y] - planeStartXInPix);
				zOfMinRowX[y] = zOfMinRowX[y] - deltaX * slopeZ;
				
				minRowX[y] = planeStartXInPix;
				}
			}
								
		for( int y=0; y<planePixHigh; y++) {		// for each row
			if( maxRowX[y] >= planeStartXInPix && minRowX[y] <= planeEndXInPix) {
				// this poly affects this row
				
				float startZ = zOfMinRowX[y];
				float deltaZ = zOfMaxRowX[y] - startZ;
				int deltaPix = maxRowX[y] - minRowX[y];
				float slopeZ = deltaZ/deltaPix;

				
				int startX = minRowX[y] + xPixArrayOffset;
				int endX = maxRowX[y] + xPixArrayOffset;
				
				float currentZ = startZ;
				for( int x=startX; x<=endX; x++) {		// for each pixel in row
					viewPlane[yOffset[y] + x].add( pID, currentZ, isBack);
					currentZ += slopeZ;
					} 
							
				}
					
			}	// end for each row					

		delete [] maxRowX;
		delete [] minRowX;
		delete [] zOfMaxRowX;
		delete [] zOfMinRowX;
		
		}	// end check for positive z vals	
	// patch has been added to zbuffer
		
	delete projected;
					
						
	} // end of zbufferpatch function